

/* bit_decoder.c
 * by Clemens Helfmeier
 *
 * This code converts a pulse-length modulated binary stream
 * into digital words.
 */

#include "bit_decoder.h"

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

/* definition for io-pin */
#define bdec_BIT		(1<<2)	// war6
#define bdec_PORT		PORTD
#define bdec_PIN		PIND
#define bdec_DIR		DDRD

/** This SAF (Signal Angepasstes Filter) is an integrator.
 * The input values are saved into an array. The integrator
 * keeps track of the integral over all values in the filter.
 * It also counts the number of times, the maximum of the integral
 * is reached (+-3).
 * The decoding routine then must be called regularily. It
 * checks for the Integral to be less than 1/4 of the Maximum
 * of the Integral and then examines if the maximum value reach
 * counter. If the counter has a high value, the maximum was quite
 * often reached an thus, the imupls was short (it didn't fill the
 * whole buffer). If the impuls was long, it filled the buffer and
 * thus has only a very view points where it was near the maximum.
 *
 * The filter length should be chosen so that the long impulses just
 * fill the filter. Then the integral will be an triangle on long
 * impulses (only once at it's maximum), and the short impulses
 * will result in a trapez (about half the filter length at its
 * maximum).
 *
 * @author Clemens Helfmeier
 * @date 12.8.06
 */

// the Length of the Filter (SAF) in multiples
// of 8.
#define FILTER_LEN		18
uint8_t bdec_F[ FILTER_LEN ];
// the current position in the array. (0..FILTER_LEN-1)
uint8_t bdec_Fai = 0;
// the currently selected bit in the filter (1<<{0..7})
uint8_t bdec_Fmsk = 1;
/* the integral over the filter values */
uint8_t bdec_Int = 0;
uint8_t bdec_IntMax = 0;
uint8_t bdec_IntCnt = 0;


/* get one bit from the stream.
 * return values are:
 * 0x00 no bit in stream available
 * 0x01 sync bit read
 * 0x02 "0" read
 * 0x03 "1" read
 * 0xF0 overflow occured (bits lost)
 */
uint8_t bdec_getBit( void ) {
	if ((bdec_Int < bdec_IntMax/4) && (bdec_IntCnt > 0)) {
		/* if the integral is reduced down to 1/4 of its maximum,
		examine if it was a 0 or a 1 */
		uint8_t ret = 0;
		/* to decide wheather the result is zero or one, we can examine the two values
		of IntCnt and IntMax. A high IntCnt or low IntMax means a short impulse, a
		low IntCnt and high IntMax means long impulses.
		*/
		/*
		uint8_t cdiff = bdec_IntCnt - (((FILTER_LEN * 8) / 2 - 6) / 2 + 6);
		uint8_t idiff = bdec_IntMax - ((FILTER_LEN * 8) * 3 / 4);
		if (cdiff - idiff < 0)
		*/
		if (bdec_IntCnt - bdec_IntMax + (-(((FILTER_LEN * 8) / 2 - 6) / 2 + 6) + ((FILTER_LEN * 8) * 3 / 4)) < 0)
			ret = bdec_ONE;
		else
			ret = bdec_ZERO;
		// reset the counter and leave the max value at its value
		// so this case won't be taken again.
		bdec_IntCnt = 0;
		return ret;
	}
	return bdec_EMPTY;
}


/* initialize io-ports
 */
void bdec_init( void ) {
	/* switch to input */
	bdec_DIR &= ~bdec_BIT;
	/* enable pull-up */
	bdec_PORT |= bdec_BIT;
}

/* this procedure must be called regularily
 */
void bdec_timer( void ) {
	/* put one bit from the port into the filter
	*/
	// advance the pointer in the array
	bdec_Fmsk = bdec_Fmsk << 1;
	if (bdec_Fmsk == 0) {
		bdec_Fmsk = 1;
		if (++bdec_Fai >= FILTER_LEN)
			bdec_Fai =0;
	}

	uint8_t F = bdec_F[ bdec_Fai ];
	// subtract the old value from the integral
	if ((F & bdec_Fmsk) != 0) bdec_Int--;
	// input the new value into the buffer and integral
	if ((bdec_PIN & bdec_BIT) == 0) {
		F |= bdec_Fmsk;
		bdec_Int++;
	} else {
		F &= ~bdec_Fmsk;
	}
	bdec_F[ bdec_Fai ] = F;
	// count the number of times, the integral is it at's maximum
	if ((bdec_IntMax - bdec_Int) <= 3)
		bdec_IntCnt++;
	// configure the maximum of the integral, this can only be one above the previous
	// maximum.
	if ((bdec_Int == bdec_IntMax + 1)) {
		bdec_IntMax = bdec_Int;
		bdec_IntCnt = 1;
	}
	// if the impuls was handled (Cnt == 0) and the
	// Integral is very low (3), reset the Maximum to the minimum value
	// that is not recognized by the recieverroutine.
	if ((bdec_IntCnt == 0) && (bdec_Int <= 3)) {
		bdec_IntMax = 3;
	}
}



